#ifdef _WINDOWS
#ifndef WINDOWS
#define WINDOWS
#endif
#endif

#include "ac_plugin.h"
#include <stdio.h>
#include <string.h>
#include <math.h>

//---------------------------------------------------------------------------------------
static void ddhcswap_calcfacenormal( Vertex *v1, Vertex *v2, Vertex *v3, Vertex *outN )
{
   Vertex nv1;
   nv1.x = v1->x - v2->x;
   nv1.y = v1->y - v2->y;
   nv1.z = v1->z - v2->z;
   Vertex nv2;
   nv2.x = v1->x - v3->x;
   nv2.y = v1->y - v3->y;
   nv2.z = v1->z - v3->z;
   CROSSPRODUCT( &nv1, &nv2, outN );
   // Normalize
   float veclen = sqrtf( outN->x * outN->x + outN->y * outN->y + outN->z * outN->z );
   if( veclen != 0 )
   {
      outN->x /= veclen;
      outN->y /= veclen;
      outN->z /= veclen;
   }
}
//---------------------------------------------------------------------------------------
// Get orientation between a normal and a location
static int ddhcswap_get_orientation( Surface *s, Vertex *v )
{
   if( s->numvert < 3 )
   {
      return 0;
   }
   Vertex *v1 = (( SVertex * )( s->vertlist->data ))->v;
   Vertex *v2 = (( SVertex * )( s->vertlist->next->data ))->v;
   Vertex *v3 = (( SVertex * )( s->vertlist->next->next->data ))->v;
   Vertex normal;
   ddhcswap_calcfacenormal( v1, v2, v3, &normal );
   float angle = DOTPRODUCT( v, &normal );
   if( angle < 0.0f )
   {
      return 1;
   }
   else if( angle > 0.0f )
   {
      return 2;
   }
   return 0;
}
//---------------------------------------------------------------------------------------
// Move/rotate objects
static void ddhcswap_go( int argc, const char* argv[] )
{

   if( argc < 7 )
   {
      message_dialog( "Insufficient parameters passed to coordswap plugin" );
      return;
   }

   // Get swap axes and negate booleans
   int swapx, swapy, swapz;
   bool negx, negy, negz;
   swapx = atoi( argv[1] );
   swapy = atoi( argv[2] );
   swapz = atoi( argv[3] );
   negx = ( atoi( argv[4] ) != 0 );
   negy = ( atoi( argv[5] ) != 0 );
   negz = ( atoi( argv[6] ) != 0 );

   if( swapx == swapy || swapx == swapz || swapy == swapz )
   {
      message_dialog( "coordswap : please ensure X, Y, and Z coordinates are not mapped to the same axis" );
      return;
   }

   // Grab all faces and vertices
   List *allsurfs = ac_selection_get_whole_surfaces_all();
   List *allverts = ac_selection_get_vertices_all();
   // Grab face normal and a vertex
   // Determine orientation of face versus vertex
   Surface *surf = NULL;
   int faceorientation = 0;
   Vertex *vert = NULL;
   Vertex pos0;
   pos0.x = pos0.y = pos0.z = 0.0f;
   if( allsurfs != NULL )
   {
      surf = ( Surface * )( allsurfs->data );
      for( List *pv = allverts; pv != NULL; pv = pv->next )
      {
         faceorientation = ddhcswap_get_orientation( surf, ( Vertex * )( pv->data ));
         if( faceorientation != 0 )
         {
            vert = ( Vertex * )( pv->data );
            break;
         }
      }
      // Realisticallym the last test should have caught the orientation.  Handle any odd geometry here, though,
      // by comparing the face normal to the origin.
      if( faceorientation == 0 )
      {
         for( List *ps = allsurfs; ps != NULL; ps = ps->next )
         {
            faceorientation = ddhcswap_get_orientation(( Surface * )( ps->data ), &pos0 );
            if( faceorientation != 0 )
            {
               surf = ( Surface * )( ps->data );
               break;
            }
         }
      }
      // If we couldn't determine an orientation based on the above tests, then we have no clear way to
      // determine whether any faces were flipped.
   }

   // Swap for selection
   add_undoable_vertex_positions( "coordinate swap", allverts );
   float vsrc[3], vdst[3];
   for( List *pv = allverts; pv != NULL; pv = pv->next )
   {
      Vertex *v = ( Vertex * )( pv->data );
      vsrc[0] = v->x;
      vsrc[1] = v->y;
      vsrc[2] = v->z;
      vdst[0] = vsrc[swapx];
      vdst[1] = vsrc[swapy];
      vdst[2] = vsrc[swapz];
      if( negx ) vdst[0] = -( vdst[0] );
      if( negy ) vdst[1] = -( vdst[1] );
      if( negz ) vdst[2] = -( vdst[2] );
      v->x = vdst[0];
      v->y = vdst[1];
      v->z = vdst[2];
   }
   list_free( &allverts );

   // Determine whether we've reversed normals
   bool flipped = false;
   if( faceorientation != 0 )
   {
      if( vert != NULL )
      {
         if( faceorientation != ddhcswap_get_orientation( surf, vert ))
         {
            // Flip all faces
            tcl_command( "ac3d surface_flip_normals" );
            //message_dialog( "Normals flipped" );
            flipped = true;
         }
      }
      else
      {
         if( faceorientation != ddhcswap_get_orientation( surf, &pos0 ))
         {
            // Flip all faces
            tcl_command( "ac3d surface_flip_normals" );
            //message_dialog( "Normals flipped" );
            flipped = true;
         }
      }
   }

   if( !flipped )
   {
      // Recalculate normals
      selected_calc_normals_force();
   }

   // Cleanup surfaces
   list_free( &allsurfs );

   find_bounding_box();
   redraw_all();
   display_status();

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      char *path = ( char * )strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/coordswap.tcl\"", path );
      ac_add_tools_menu_item( "Coordinate swap...", "ddhcswap_dialog",
                              "Swap vertex axes to support different coordinate systems" );
      ac_add_command_full( "ddhcswap_go", ( void * )ddhcswap_go,
                            0, "argv", "<arg1> <arg2>",
                           "Swap vertex axes to support different coordinate systems" );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Coordinate swap - version 1.0 - Dennis Hawthorne" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Coordinate swap - Dennis Hawthorne - Swap selected vertices' X/Y/Z coordinates, for the purposes of flipping geometry or converting to a different coordinate system." );
}
//---------------------------------------------------------------------------------------
